#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Geländemodellierung mit AutoCAD oder BricsCAD in drei Schritten
#
# 1. Ermittlung von 3D-Daten aus einer xyh-Datei von www.gpsvisualizer.com
# 2. Erstellung eines Dreiecksnetzes
# 3. Konstruktion eines Dreicksprismenfeldes in AutoCAD oder BricsCAD, das
#    durch das Dreiecksnetz horizontal geteilt wird.
#
# Dipl.-Ing. Martin Vogel, Hochschule Bochum
# Lizenz: CC-by-3.0 – https://creativecommons.org/licenses/by/3.0/de/
# Version vom 2016-06-08
#
# ToDo: Das Programm findet momentan die Dreiecke mit den kürzesten
# Seitenlängen.
# Bei einem Viereck aus zwei Dreiecken mit gemeinsamer Kante liegt diese Kante
# immer auf der kürzeren der beiden Diagonalen. Das kann zu Dreiecken mit sehr
# kleinen Innenwinkeln und damit sehr kleiner Fläche neben deutlich größeren
# Nachbarn führen.
# Ein Tausch der Diagonalen würde in so einem Fall zu ausgewogeneren Netzen
# mit gleichmäßigerer Maschengröße und weniger spitzwinkligen Dreiecken führen.
# Lohnt sich der Aufwand?
#
# ToDoToo: Noch sinnvoller als eine Scriptdatei wäre ja eine DXF-Datei.
# Wenn Autodesk sich nur nicht so erbärmlich anstellte und 3DSOLIDs zum
# Firmengeheimnis erklärte!
# http://www.autodesk.com/techpubs/autocad/acad2000/dxf/3dsolid_dxf_06.htm
# "Proprietary data (multiple lines < 255 characters each)"
# Sowas passiert, wenn man keine offenen Formate verwendet, sondern gierig
# obskure Lizenzen einkauft, wie hier die ACIS-Datenstruktur von Spatial
# Technologies.
# http://www.crlf.de/Dokumente/Dxf-R13/DXF13ENT.HTM
# Wenigstens ist die Verschlüsselung nur pro forma:
# https://www.c-plusplus.net/forum/255950-full
# Das ACIS-Format selbst wird hier beschrieben:
# http://paulbourke.net/dataformats/sat/sat.pdf

from sys import exit as SysExit
from math import cos, radians, hypot
from tkinter import Tk, filedialog, Canvas, Label, messagebox, mainloop

##############################################################
# Einleitung: Programmvorstellung, Auswahl der Eingangsdatei
#
# Koordinaten-Datei suchen …
Fenster = Tk()
Fenster.title("Python-Triangulation")
Canvasbreite = 800
Canvashöhe = 600
C = Canvas(Fenster, width=Canvasbreite, height=Canvashöhe, bg="white")
C.grid(row=0, column=0, columnspan=2)
Statuszeile = Label(text="Bitte, wählen Sie die im zweiten Schritt erzeugte"
                    " TXT-Datei jetzt aus!", font="sansserif 11 bold")
Statuszeile.grid(row=1, column=0, columnspan=2)
PnL = Label(text="./.")
PnL.grid(row=2, column=0, sticky="e")
Label(text="Koordinatenpunkte in der TXT-Datei gefunden."
      ).grid(row=2, column=1, sticky="w")
LnL = Label(text="./.")
LnL.grid(row=3, column=0, sticky="e")
Label(text="mögliche Verbindungslinen erzeugt und nach Länge sortiert."
      ).grid(row=3, column=1, sticky="w")
KnL = Label(text="./.")
KnL.grid(row=4, column=0, sticky="e")
Label(text="sich nicht gegenseitig schneidende Kanten gefunden."
      ).grid(row=4, column=1, sticky="w")
DnL = Label(text="./.")
DnL.grid(row=5, column=0, sticky="e")
Label(text="Dreiecke gebildet."
      ).grid(row=5, column=1, sticky="w")


def Status(t):
    Statuszeile.config(text=t)
    # Statuszeile.update()


def Fehler(t, detail=""):
    messagebox.showerror("Fehler", t, detail=detail)
    Fenster.destroy()
    SysExit()

Vorstellungstext = """
Dieses Programm hilft bei der Erzeugung von Volumenkörpern aus 3D-Daten,
indem es eine Scriptdatei schreibt, die von AutoCAD oder BricsCAD
mit dem Befehl SCRIPT ausgeführt werden kann.

Um eine geeignete Eingabedatei zu erhalten, suchen Sie auf
http://www.gpsvisualizer.com/draw/ ein Geländestück und zeichnen Sie dort eine
kreuz und quer laufende Strecke mit allen interessanten Geländepunkten auf.
Die Punkte sollten nicht dichter als 30 Meter nebeneinander liegen, weil die
vorhandenen Daten aus der Radarhöhenmessung des SRTM-Projektes ohnehin nicht
feiner aufgelöst sind.

Exportieren Sie die Strecke, die zwischen einigen dutzend und einigen hundert
Punkten besitzen sollte, als Datei.

Laden Sie diese Datei auf http://www.gpsvisualizer.com/elevation
hoch, lassen Sie die Höhen dazurechnen und exportieren Sie die Ergebnisse
anschließend als Datei im TXT-Format!

Diese neue Datei besteht nun aus zahlreichen Zeilen wie z.B.
T	51.445448500	7.269880800	129.8
Die tabulatorgetrennten Werte hinter dem "T" sind der Breitengrad,
der Längengrad und die Höhe über NN jedes Geländepunktes.
"""
C.create_text(Canvasbreite/2, Canvashöhe/2,
              text=Vorstellungstext,
              font="sansserif 12")
xyhname = filedialog.askopenfilename(
    parent=Fenster,
    filetypes=(("Textdateien", "*.txt"),
               ("Alle Dateien", "*.*")),
    initialfile="GPSVisualizer.txt")

##############################################################
# 1. Akt: Einlesen der Längen-, Breiten- und Höhenwerte
#
Status("Lese "+xyhname+" …")
xyh = open(xyhname, "r")
try:
    k = xyh.read()
except:
    Fehler("Diese Datei kann nicht gelesen werden.",
           detail="Enthält sie wirklich lesbaren Text?")
xyh.close()

# Uns interessiert nur der Abschnitt mit den Koordinaten.
# LatLonH ist eine Liste aus Tupeln (Längengrad, Breitengrad, Höhe)
LatLonH = []
Warnung = False
for Zeile in k.split("\n"):
    if len(Zeile) > 1 and Zeile[0] == "T":  # Das "T" kennzeichnet Zeilen mit
                                            # trigonometrischen Daten
        try:
            LatLonH.append([eval(x) for x in Zeile.split("\t")[1:4]])
        except:
            Warnung = True

if Warnung:
    Status("Warnung: Mindestens eine Zeile der Datei wurde nicht erkannt!")

Punktzahl = len(LatLonH)
PnL.config(text=Punktzahl)
if Punktzahl < 3:
    Fehler("Nicht genug Punkte in der Eingabedatei gefunden.",
           detail="Tut mir leid, aber für ein Dreieck brauche ich "
           "mindestens drei Punkte. Sie verstehen.")

# Um Trapezverzerrungen zu vermeiden, wird der Mittelpunkt des kartierten
# Bereichs gesucht.
# Dieser wird als Referenzpunkt zur Umrechnung von Grad auf Meter verwendet.
# Alle anderen Punkte verwenden denselben Maßstab (konforme Projektion).
Lat_Werte = [LatLonH[i][1] for i in range(Punktzahl)]
Lat_Mitte = (min(Lat_Werte)+max(Lat_Werte))/2
Lon_Werte = [LatLonH[i][0] for i in range(Punktzahl)]
Lon_Mitte = (min(Lon_Werte)+max(Lon_Werte))/2
Cos_Lon_Mitte = cos(radians(Lon_Mitte))

# Näherungsweise Umrechnung von Gradkoordinaten auf Meter.
# Grob vereinfachend wird die Erde als Kugel mit dem Umfang 40000000 Meter
# angenommen.
# Natürlich kann man das exakter rechnen, dann sollte man aber auch keine
# SRTM-Höhenwerte aus Onlinediensten zusammenklicken.
eu = 40000000                           # 40 Millionen Meter.
try:
    Gm = [(p[1]*eu/360*Cos_Lon_Mitte,   # Ost-West-Abstand der Längenkreise
                                        # nimmt zum Pol hin ab.
           p[0]*eu/360,                 # Abstand der Breitenkreise bleibt
                                        # gleich.
           p[2]                         # Höhe unverändert in Metern.
           ) for p in LatLonH]          # … und das für alle Punkte der Datei.
except:
    Fehler("Zahlenwerte unvollständig",
           detail="Fehlen vielleicht die Höhenwerte?")


# Koordinate der linken unteren Ecke – wird später (0,0).
x_min = min(xyh[0] for xyh in Gm)
y_min = min(xyh[1] for xyh in Gm)

# Verschieben aller Koordinaten in Richtung Ursprung:
# Zum einen werden die Koordinaten dann lesbarer (die Entfernung in Metern
# vom Äquator bzw. Nullmeridian ist eine recht unhandliche Zahl),
# zum anderen umgehen wir einen Fehler im Kappen-Befehl von BricsCAD 15,
# der dafür sorgt, dass der Befehl nur "in der Nähe" des Ursprungs
# funktioniert.
Gm = [[xyh[0]-x_min, xyh[1]-y_min, xyh[2]] for xyh in Gm]

# Grenzen des interessanten Quaders
x_min = 0
x_max = max(xyh[0] for xyh in Gm)
y_min = 0
y_max = max(xyh[1] for xyh in Gm)
h_min = min(xyh[2] for xyh in Gm)
h_max = max(xyh[2] for xyh in Gm)

##############################################################
# 2. Akt: Triangulation des Punktefeldes zu einem Dreiecksnetz
#
# Eingang: Liste von Punkten Gm (nur x- und y-Ordinate werden verwendet)
# Daraus wird eine nach Abstand sortierte Liste aller Verbindungsmöglichkeiten
# erzeugt.
# Nun werden, beginnend bei der kürzesten Verbindung, alle Verbindungen
# gesucht, die nicht zu einer Kollison führen.
# Daraus werden schließlich Dreiecke gebildet.
Status("Suche nach Kanten für das Dreiecksnetz …")

Punkte = Gm     # Liste aller Punktkoordinaten (in Metern)
Linien = []     # Liste aller(!) möglichen Punkteverbindungen
Kanten = []     # Liste aller verwendeten Punkteverbindungen
Dreiecke = []   # Liste aller Dreiecke

# Canvaskoordinaten eines Punktes
# (isometrisch um Höhen erweitert)
Cx = Canvasbreite/((x_max-x_min)+(y_max-y_min)*0.35)
Cy = Canvashöhe/((y_max-y_min)*0.35+(h_max-h_min))


def CK(x, y, h=0):
    return (x+y*0.35)*Cx, Canvashöhe-(y*0.35+h-h_min)*Cy


def ZeichnePunkt(P):
    x, y = CK(*P)
    x, y_min = CK(P[0], P[1], h_min)
    # Punkt in Originalhöhe
    C.create_oval(x-1, y-1, x+1, y+1, fill="blue", outline="blue")
    # Punkt auf Basishöhe h_min
    C.create_oval(x-1, y_min-1, x+1, y_min+1, fill="red", outline="red")
    # Vertikale Verbindungslinie zwischen den beiden Punkten
    C.create_line(x, y, x, y_min, fill="black")


def ZeichneLinie(L, color="black"):
    C.create_line(CK(*Punkte[L[0]]), CK(*Punkte[L[1]]), fill=color)


def ZeichneDreieck(D):
    # Dreieck in Originalhöhe
    C.create_polygon(CK(*Punkte[D[0]]),
                     CK(*Punkte[D[1]]),
                     CK(*Punkte[D[2]]),
                     outline="blue", fill="")
    # Dreieck auf h_min
    C.create_polygon(CK(Punkte[D[0]][0], Punkte[D[0]][1], h_min),
                     CK(Punkte[D[1]][0], Punkte[D[1]][1], h_min),
                     CK(Punkte[D[2]][0], Punkte[D[2]][1], h_min),
                     outline="red", fill="")


def Neuzeichnen():
    C.delete("all")
    for P in Punkte:
        ZeichnePunkt(P)
#    for L in Linien:
#        ZeichneLinie(L, "gray")
#    for K in Kanten:
#        ZeichneLinie(K, "green")
    for D in Dreiecke:
        ZeichneDreieck(D)
    C.update()


def Überschneidung(L1, L2):
    # Schneiden sich zwei Strecken zwischen ihren Endpunkten?
    # Beschreibung des Algorithmus auf http://stackoverflow.com
    # /questions/563198
    # /how-do-you-detect-where-two-line-segments-intersect

    # Gibt es einen gemeinsamen Endpunkt?
    # Das zählt bei diesem Problem nicht als Schnitt.
    if L1[0] in L2 or L1[1] in L2:
        return False

    # Für mehr Lesbarkeit:
    p0_x, p0_y, z = Punkte[L1[0]]
    p1_x, p1_y, z = Punkte[L1[1]]
    p2_x, p2_y, z = Punkte[L2[0]]
    p3_x, p3_y, z = Punkte[L2[1]]

    # Höhen- und Breitendifferenzen der beiden Strecken
    s10_x = p1_x - p0_x
    s10_y = p1_y - p0_y
    s32_x = p3_x - p2_x
    s32_y = p3_y - p2_y

    Nenner = s10_x * s32_y - s32_x * s10_y
    if Nenner == 0:
        return False  # Kollinear

    NennerPositiv = Nenner > 0

    # Höhen- und Breitendifferenz der Linie vom ersten zum letzten Punkt
    s02_x = p0_x - p2_x
    s02_y = p0_y - p2_y

    s_numer = s10_x * s02_y - s10_y * s02_x
    if (s_numer < 0) == NennerPositiv:
        return False  # Kein Schnitt

    t_numer = s32_x * s02_y - s32_y * s02_x
    if (t_numer < 0) == NennerPositiv:
        return False  # Kein Schnitt

    if ((s_numer > Nenner) == NennerPositiv) or\
       ((t_numer > Nenner) == NennerPositiv):
        return False  # Kein Schnitt
    # Jetzt bleibt nur noch die Überschneidung übrig.
    return True

Pn = len(Punkte)
Neuzeichnen()

# Jetzt stellen wir alle möglichen Verbindungslinien zwischen jeweils
# zwei Punkten zusammen.
for i in range(Pn-1):
    for j in range(i+1, Pn):
        Linien.append((i, j))
        #  ZeichneLinie((i,j),"lightgray")
Ln = len(Linien)  # Das sollten jetzt (Pn**2-Pn)/2 Linien sein.


# Linien nach Länge sortieren:
def Linienlänge(L):
    return hypot(Punkte[L[0]][0]-Punkte[L[1]][0],
                 Punkte[L[0]][1]-Punkte[L[1]][1])
Linien = sorted(Linien, key=Linienlänge)
LnL.config(text=Ln)

# Die Linien, die sich nicht mit anderen Linien schneiden, werden nun
# zu Dreieckskanten. Da wir mit der kürzesten Linie beginnen und
# die Linien nach Länge sortiert sind, kommt ein ganz hübsches
# Dreiecksnetz heraus.
for L in Linien:
    for K in Kanten:
        if Überschneidung(L, K):
            break  # Weiter müssen wir nicht suchen
    else:  # Python <3 :-)
        Kanten.append(L)
        ZeichneLinie(L, color="green")
        C.update()
Kn = len(Kanten)
KnL.config(text=Kn)

Status("Suche nach geeigneten Dreiecken …")


# Jetzt die Dreiecke einsammeln:
# Für alle Kanten K von P1 nach P2:
#   Für alle Punkte P
#     Gibt es eine Kante P-P1 und eine Kante P-P2?
#       Dreieck P-P1-P2 in Dreiecksliste aufnehmen, falls
#       es noch nicht existiert (sinnvollerweise sind P, P1 und P2
#       für jedes Dreieck sortiert)
# Um zu vermeiden, dass ein großes Dreieck mehrere kleine umfasst,
# werden nur Dreiecke zugelassen, die keine weiteren Punkte des Netzes
# einschließen.
def PunktInDreieck(P, P1, P2, P3):
    # http://totologic.blogspot.de/2014/01/accurate-point-in-triangle-test.html
    Nenner = ((P2[1]-P3[1])*(P1[0]-P3[0])+(P3[0]-P2[0])*(P1[1]-P3[1]))
    if not Nenner:
        return False  # Dreieck der Fläche null
    a = ((P2[1]-P3[1])*(P[0]-P3[0])+(P3[0]-P2[0])*(P[1]-P3[1]))/Nenner
    b = ((P3[1]-P1[1])*(P[0]-P3[0])+(P1[0]-P3[0])*(P[1]-P3[1]))/Nenner
    c = 1 - a - b
    return 0 <= a <= 1 and 0 <= b <= 1 and 0 <= c <= 1

Dv = 0  # Zähler für verworfene Dreiecke, die aus mehreren kleinen bestehen
for K in Kanten:  # Für alle Kanten:
    for i in range(Pn):  # Für alle Punkte …
        if i not in K:   # … die nicht zur aktuellen Kante gehören:
            # Um eine Kante zu finden, müssen die Punktindizes
            # aufsteigend sortiert sein. Es kann zwar eine Kante von
            # Punkt 4 nach Punkt 17 geben, jedoch niemals eine Kante
            # von Punkt 17 nach Punkt 4.
            a, b = sorted([i, K[0]])
            c, d = sorted([i, K[1]])
            if (a, b) in Kanten and (c, d) in Kanten:  # Punkt gefunden!
                # Auch im Dreieck müssen die Punktindizes stets
                # aufsteigend sortiert sein, um Eindeutigkeit zu erreichen.
                D = sorted([i, K[0], K[1]])
                if D not in Dreiecke:    # Möglicher Kandidat gefunden.
                    for i in range(Pn):  # Alle Punkte prüfen, nur …
                        if i not in D:   # … die eigenen Eckpunkte nicht.
                            if PunktInDreieck(Punkte[i], Punkte[D[0]],
                                              Punkte[D[1]], Punkte[D[2]]):
                                Dv += 1  # Dieses Dreieck ist überflüssig.
                                break
                    else:  # neues Dreieck gefunden
                        Dreiecke.append(D)
                        ZeichneDreieck(D)
                        C.update()
Dn = len(Dreiecke)
DnL.config(text=Dn)

# Vorschaubild anzeigen
Neuzeichnen()

#############################################################
# 3. und letzter Akt: Export der Dreiecke als CAD-Skriptdatei

Status("Schreibe BricsCAD/AutoCAD-Skripdatei …")

# Anlegen der Skriptdatei.
scr = open(xyhname+".scr", "w")


# Funktion, um eine Zeile der Scriptdatei zu schreiben.
# Wer mag, kann das so umbauen, dass die Befehle direkt an AutoCAD
# geschickt werden. Siehe http://code.google.com/p/pywinauto
def acad(zeile):
    print(zeile, file=scr)

# Koordinateneingabe bekommt Vorrang vor Objektfang.
acad("OSnapCoord 1")

# Isometrische Ansicht wählen.
acad("APunkt 1,1,1")

# Hilfsquader zum Zoomen auf den interessanten Bereich zeichnen.
# Der Quader wird nach dem Zoomen gelöscht.
Höhe = h_max-h_min
acad("Quader %f,%f,%f %f,%f,%f" % (x_min, y_min, h_min,
                                       x_max, y_max, h_max))
acad("Zoom G")
acad("_Erase L ")

# Der Geländekörper ist doppelt so hoch wie die maximale Höhendifferenz.
# Unten und oben reicht er jeweils um die halbe Höhe über die
# Geländegrenzen hinaus.
# Zu schade, dass der Befehl "Erhebung" in AutoCAD 2016 nicht mehr ordentlich
# in Scripten funktioniert, jetzt müssen die Basisdreiecke mit 3DPOLY statt
# PLINIE konstruiert werden.
Basishöhe = h_min-0.5*Höhe
# acad("Erhebung %f 0"%(Basishöhe))

# Für jedes Dreieck aus der Triangulation die Koordinaten heraussuchen:
for P in [(Punkte[D[0]], Punkte[D[1]], Punkte[D[2]]) for D in Dreiecke]:
    # Grundfläche zeichnen …
    acad("3DPOLY %f,%f,%f %f,%f,%f %f,%f,%f s" %
         (P[0][0], P[0][1], Basishöhe,
          P[1][0], P[1][1], Basishöhe,
          P[2][0], P[2][1], Basishöhe))
    # … und um das doppelte der maximalen Höhe extrudieren.
    # "L" bezeichnet in AutoCAD und BricsCAD das zuletzt gezeichnete Objekt.
    acad("_extrude L  %f" % (2*Höhe))
    # Kappebene durch die drei Punkte der räumlichen Dreiecksfläche ziehen.
    # Die Anzahl der Leerzeichen nach dem L ist wichtig: 3!
    acad("KAPPEN L   %f,%f,%f %f,%f,%f %f,%f,%f " %
         (P[0][0], P[0][1], P[0][2],
          P[1][0], P[1][1], P[1][2],
          P[2][0], P[2][1], P[2][2]))

# Schließen der Scriptdatei
scr.close()

# Schlussmeldung, Statistik
Status("Fertig! "+xyhname+".scr wurde angelegt.")
mainloop()
